import networkx as nx

from abc import ABC, abstractmethod
from typing import TypeVar, Generic
from lxml import etree as ET

from xgmml.utils.xml_utils import *


T = TypeVar('T')


class XGMMLReader(Generic[T], ABC):
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def _networkx_to(self, graph: nx.MultiGraph, **kwargs) -> T:
        raise NotImplementedError()

    def _read_graph(
        self, graph_xml: ET.ElementTree, namespaces: dict, **kwargs
    ) -> nx.MultiGraph:
        graph_directed = graph_xml.xpath('/g:graph/@directed', namespaces=namespaces)
        graph_directed = (
            True
            if graph_directed[0] == '1'
            else False
            if len(graph_directed) == 1
            else False
        )

        return nx.MultiDiGraph() if graph_directed else nx.MultiGraph()

    def _read_node(
        self,
        graph: nx.MultiGraph,
        graph_xml: ET.ElementTree,
        namespaces: dict,
        **kwargs,
    ):
        for node_xml in graph_xml.xpath('/g:graph/g:node', namespaces=namespaces):
            node_id = node_xml.attrib['id']

            node_attributes = {}

            for att_name, att_value in node_xml.attrib.items():
                if att_name in IGNORED_NODE_ATT_ELEMENTS:
                    continue

                node_attributes[att_name] = att_value

            for att_xml in node_xml.xpath('g:att', namespaces=namespaces):
                att_name = first_or_none(att_xml.xpath('@name'))
                att_type = first_or_none(att_xml.xpath('@type'))
                att_value = first_or_none(att_xml.xpath('@value'))

                if att_name is None or att_type is None or att_value is None:
                    continue

                if att_name in IGNORED_NODE_ATT_ELEMENTS:
                    continue

                att_value = get_att_value(att_value, value_type=att_type)

                node_attributes[att_name] = att_value

            graph.add_node(node_id, **node_attributes)

    def _read_edge(
        self,
        graph: nx.MultiGraph,
        graph_xml: ET.ElementTree,
        namespaces: dict,
        **kwargs,
    ):
        for edge_idx, edge_xml in enumerate(
            graph_xml.xpath('/g:graph/g:edge', namespaces=namespaces)
        ):
            source_node_id = edge_xml.attrib['source']
            target_node_id = edge_xml.attrib['target']

            edge_attributes = {}

            for att_name, att_value in edge_xml.attrib.items():
                if att_name in IGNORED_EDGE_ATT_ELEMENTS:
                    continue

                edge_attributes[att_name] = att_value

            for att_xml in edge_xml.xpath('g:att', namespaces=namespaces):
                att_name = first_or_none(att_xml.xpath('@name'))
                att_type = first_or_none(att_xml.xpath('@type'))
                att_value = first_or_none(att_xml.xpath('@value'))

                if att_name is None or att_type is None or att_value is None:
                    continue

                if att_name in IGNORED_EDGE_ATT_ELEMENTS:
                    continue

                att_value = get_att_value(att_value, value_type=att_type)

                edge_attributes[att_name] = att_value

            graph.add_edge(
                source_node_id, target_node_id, key=edge_idx, **edge_attributes
            )

    def read(self, xgmml_path: str, **kwargs) -> T:
        graph_xml = ET.parse(xgmml_path, ET.XMLParser())

        namespaces = {}
        for prefix, uri in NAMESPACES.items():
            if prefix is not None:
                namespaces[prefix] = uri
            else:
                namespaces['g'] = uri

        graph = self._read_graph(graph_xml, namespaces, **kwargs)
        self._read_node(graph, graph_xml, namespaces, **kwargs)
        self._read_edge(graph, graph_xml, namespaces, **kwargs)

        return self._networkx_to(graph, **kwargs)


__all__ = ['XGMMLReader']



import networkx as nx

from xgmml.reader.xgmml_reader import XGMMLReader


class NetworkxXGMMLReader(XGMMLReader[nx.MultiGraph]):
    def __init__(self, **kwargs):
        super(NetworkxXGMMLReader, self).__init__(**kwargs)

    def _networkx_to(self, graph: nx.MultiGraph, **kwargs) -> nx.MultiGraph:
        return graph


__all__ = ['NetworkxXGMMLReader']



import igraph as ig
import networkx as nx

from xgmml.layout.base_layout import Layout


class KamadaKawaiLayout(Layout):
    def __init__(self, **kwargs):
        super(KamadaKawaiLayout, self).__init__('Kamada-Kawai Layout', **kwargs)

    def _subgraph_layout(
        self, subgraph: nx.Graph, nodes_size: dict[str, tuple[float, float]] = None
    ) -> dict[str, tuple[float, float]]:
        ig_subgraph: ig.Graph = ig.Graph.from_networkx(subgraph)
        ig_subgraph_layout: ig.Layout = ig_subgraph.layout_kamada_kawai()

        subgraph_layout = {}
        for node_id, node_position in zip(subgraph.nodes, ig_subgraph_layout.coords):
            subgraph_layout[node_id] = node_position

        scale = self._calc_auto_scale(subgraph, subgraph_layout, nodes_size)

        return nx.rescale_layout_dict(subgraph_layout, scale=scale)


__all__ = ['KamadaKawaiLayout']



import uuid
import networkx as nx

from abc import ABC, abstractmethod
from typing import TypeVar, Generic
from datetime import datetime
from lxml import etree as ET
from lxml.builder import ElementMaker, E
from lxml.etree import QName

from xgmml.layout.base_layout import Layout
from xgmml.layout.kamada_kawai_layout import KamadaKawaiLayout
from xgmml.utils.xml_utils import *


T = TypeVar('T')


class XGMMLWriter(Generic[T], ABC):
    def __init__(self, **kwargs):
        self._uuid = str(uuid.uuid4())
        self._element_maker = ElementMaker(nsmap=NAMESPACES)

    @abstractmethod
    def _to_networkx(self, graph: T, **kwargs) -> nx.MultiGraph:
        raise NotImplementedError()

    def _create_value(
        self,
        att_name: str,
        att_value: dict,
        source_data=None,
        source_data_dict=None,
        default_value=None,
        **kwargs,
    ):
        kwarg = att_value['kwarg']
        default = att_value['default']
        value = kwargs.get(kwarg, default)
        value_type = att_value['type'] or get_att_value_type(value)

        mapper_kwarg = att_value.get('mapper_kwarg', None)
        key_kwarg = att_value.get('key_kwarg', None)
        mapper_key = kwargs.get(key_kwarg, None)

        if mapper_kwarg is not None and key_kwarg is not None:
            mapper_cls = kwargs.get(mapper_kwarg, att_value.get('default_mapper', None))
            source_value = source_data.get(mapper_key, None)
            source_values = source_data_dict.get(mapper_key, None)

            if (
                mapper_cls is not None
                and source_value is not None
                and source_values is not None
            ):
                mapper = mapper_cls(source_values, **kwargs)
                value = mapper.map(source_value)

        if value is None:
            value = default_value

        return create_att_value(value, value_type=value_type)

    def _create_att_list(
        self,
        atts: dict[str, dict],
        source_data: dict = None,
        source_data_dict: dict[str, list] = None,
        **kwargs,
    ) -> dict[str, str]:
        att_dict = {}

        for att_name, att_value in atts.items():
            value = self._create_value(
                att_name,
                att_value,
                source_data=source_data,
                source_data_dict=source_data_dict,
                **kwargs,
            )

            if value is not None:
                att_dict[att_name] = value

        return att_dict

    def _create_att_xml_list(
        self,
        atts: dict[str, dict],
        source_data: dict = None,
        source_data_dict: dict[str, list] = None,
        **kwargs,
    ) -> list[ET.Element]:
        att_xml_list = []

        for att_name, att_value in atts.items():
            value_type = att_value['type']
            value = self._create_value(
                att_name,
                att_value,
                source_data=source_data,
                source_data_dict=source_data_dict,
                **kwargs,
            )

            if value is not None:
                att_xml_list.append(
                    E.att(create_att_name_value(att_name, value, value_type=value_type))
                )

        return att_xml_list

    def _create_graph_att_xml_list(
        self, graph_layout: Layout, **kwargs
    ) -> list[ET.Element]:
        graph_type = kwargs.get('graph_type', 'N/A')
        graph_description = kwargs.get('graph_description', 'N/A')
        graph_identifier = kwargs.get('graph_identifier', 'N/A')
        graph_title = kwargs.get('graph_label', self._uuid)
        graph_source = kwargs.get('graph_source', 'https://xgmml.leovan.tech')
        graph_format = kwargs.get('graph_format', 'Cytoscape-XGMML')

        att_xml_list = []

        # rdf
        rdf_element = ET.Element(
            QName(NAMESPACES['rdf'], 'RDF'),
        )

        rdf_description_element = ET.SubElement(
            rdf_element,
            QName(NAMESPACES['rdf'], 'Description'),
            {
                QName(NAMESPACES['rdf'], 'about'): 'https://xgmml.leovan.tech',
            },
        )

        rdf_dc_description_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'type')
        )
        rdf_dc_description_element.text = graph_type

        rdf_dc_description_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'description')
        )
        rdf_dc_description_element.text = graph_description

        rdf_dc_identifier_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'identifier')
        )
        rdf_dc_identifier_element.text = graph_identifier

        rdf_dc_date_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'date')
        )
        rdf_dc_date_element.text = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        rdf_dc_title_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'title')
        )
        rdf_dc_title_element.text = graph_title

        rdf_dc_source_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'source')
        )
        rdf_dc_source_element.text = graph_source

        rdf_dc_format_element = ET.SubElement(
            rdf_description_element, QName(NAMESPACES['dc'], 'format')
        )
        rdf_dc_format_element.text = graph_format

        network_metadata = E.att(
            rdf_element,
            {
                'name': 'networkMetadata',
            },
        )
        att_xml_list.append(network_metadata)

        # others
        att_xml_list.extend(
            [
                E.att(create_att_name_value('shared name', graph_title)),
                E.att(create_att_name_value('name', graph_title)),
                E.att(create_att_name_value('selected', '1', value_type='boolean')),
                E.att(create_att_name_value('layoutAlgorithm', graph_layout.name)),
            ]
        )

        return att_xml_list

    def _create_graph_graphics_xml(self, nodes_position: dict, **kwargs) -> ET.Element:
        graph_label = kwargs.get('graph_label', self._uuid)

        nodes_position_x = [position[0] for position in nodes_position.values()]
        nodes_position_x_min = min(nodes_position_x)
        nodes_position_x_max = max(nodes_position_x)
        graph_width = nodes_position_x_max - nodes_position_x_min
        graph_center_x = (nodes_position_x_max - nodes_position_x_min) / 2.0

        nodes_position_y = [position[1] for position in nodes_position.values()]
        nodes_position_y_min = min(nodes_position_y)
        nodes_position_y_max = max(nodes_position_y)
        graph_height = nodes_position_y_max - nodes_position_y_min
        graph_center_y = (nodes_position_y_max - nodes_position_y_min) / 2.0

        att_xml_list = [
            E.att(create_att_name_value('NETWORK_TITLE', graph_label)),
            E.att(create_att_name_value('NETWORK_WIDTH', graph_width)),
            E.att(create_att_name_value('NETWORK_HEIGHT', graph_height)),
            E.att(create_att_name_value('NETWORK_CENTER_X_LOCATION', graph_center_x)),
            E.att(create_att_name_value('NETWORK_CENTER_Y_LOCATION', graph_center_y)),
        ]

        att_xml_list.extend(
            self._create_att_xml_list(GRAPH_GRAPHICS_ATT_ELEMENTS, **kwargs)
        )

        graphics_element = E.graphics(*att_xml_list)

        return graphics_element

    def _create_node_att_xml_list(
        self, node_id, node_data: dict, **kwargs
    ) -> list[ET.Element]:
        att_xml_list = [
            E.att(create_att_name_value('name', node_id)),
            E.att(create_att_name_value('shared name', node_id)),
        ]

        for name, value in node_data.items():
            if name in IGNORED_NODE_ATT_ELEMENTS:
                continue

            value_type = get_att_value_type(value)
            att_xml_list.append(E.att(create_att_name_value(name, value, value_type)))

        return att_xml_list

    def _create_node_graphics_xml(
        self,
        node_data: dict,
        node_data_dict: dict[str, list],
        node_position: tuple[float, float],
        **kwargs,
    ) -> ET.Element:
        graphics_att_xml_list = self._create_att_xml_list(
            NODE_GRAPHICS_ATT_ELEMENTS, node_data, node_data_dict, **kwargs
        )

        return E.graphics(
            *graphics_att_xml_list,
            {
                **self._create_att_list(
                    NODE_GRAPHICS_ATTS, node_data, node_data_dict, **kwargs
                ),
                'x': str(node_position[0]),
                'y': str(node_position[1]),
                'z': '0.0',
            },
        )

    def _create_node_data_dict(
        self,
        graph: nx.MultiGraph,
    ) -> dict[str, list]:
        node_data_dict = {}
        for _, node_data in graph.nodes(data=True):
            for k, v in node_data.items():
                if k not in node_data_dict:
                    node_data_dict[k] = []

                node_data_dict[k].append(v)

        return node_data_dict

    def _create_node_xml_list(
        self,
        graph: nx.MultiGraph,
        nodes_position: dict[str, tuple[float, float]],
        **kwargs,
    ) -> list[ET.Element]:
        node_data_dict = self._create_node_data_dict(graph)
        node_xml_list = []

        for node_id, node_data in graph.nodes(data=True):
            node_position = nodes_position.get(node_id, (0.0, 0.0))

            att_xml_list = self._create_node_att_xml_list(node_id, node_data, **kwargs)
            att_xml_list.extend(
                self._create_att_xml_list(
                    NODE_ATT_ELEMENTS, node_data, node_data_dict, **kwargs
                )
            )

            graphics_xml = self._create_node_graphics_xml(
                node_data, node_data_dict, node_position, **kwargs
            )

            node_xml = E.node(
                *att_xml_list,
                graphics_xml,
                {
                    'id': node_id,
                    'label': self._create_value(
                        'label',
                        NODE_ATTS['label'],
                        node_data,
                        node_data_dict,
                        default_value=node_id,
                        **kwargs,
                    ),
                },
            )

            node_xml_list.append(node_xml)

        return node_xml_list

    def _create_edge_label(
        self, source_node_id, target_node_id, edge_type, edge_data: dict, **kwargs
    ) -> str:
        edge_label_key = kwargs.get('edge_label_key', None)
        edge_label = edge_data.get(edge_label_key, None)

        if edge_label is None:
            edge_label = f'{source_node_id} ({edge_type}) {target_node_id}'

        return edge_label

    def _create_edge_att_xml_list(
        self, source_node_id, target_node_id, edge_data: dict, **kwargs
    ) -> list[ET.Element]:
        edge_type_key = kwargs.get('edge_type_key', None)
        edge_type = edge_data.get(edge_type_key, '-')

        edge_label = self._create_edge_label(
            source_node_id, target_node_id, edge_type, edge_data, **kwargs
        )

        att_xml_list = [
            E.att(create_att_name_value('name', edge_label)),
            E.att(create_att_name_value('shared name', edge_label)),
            E.att(create_att_name_value('interaction', edge_type)),
            E.att(create_att_name_value('shared interaction', edge_type)),
        ]

        for name, value in edge_data.items():
            if name in IGNORED_EDGE_ATT_ELEMENTS:
                continue

            value_type = get_att_value_type(value)
            att_xml_list.append(E.att(create_att_name_value(name, value, value_type)))

        return att_xml_list

    def _create_edge_graphics_xml(
        self, edge_data: dict, edge_data_dict: dict[str, list], **kwargs
    ) -> ET.Element:
        graphics_att_xml_list = []

        graphics_att_xml_list.extend(
            self._create_att_xml_list(
                EDGE_GRAPHICS_ATT_ELEMENTS, edge_data, edge_data_dict, **kwargs
            )
        )

        return E.graphics(
            *graphics_att_xml_list,
            {
                'width': self._create_value(
                    'width',
                    EDGE_GRAPHICS_ATTS['width'],
                    edge_data,
                    edge_data_dict,
                    **kwargs,
                ),
                'fill': self._create_value(
                    'fill',
                    EDGE_GRAPHICS_ATTS['fill'],
                    edge_data,
                    edge_data_dict,
                    **kwargs,
                ),
            },
        )

    def _create_edge_xml_list(self, graph: nx.MultiGraph, **kwargs) -> list[ET.Element]:
        edge_data_dict = {}
        for _, _, edge_data in graph.edges(data=True):
            for k, v in edge_data.items():
                if k not in edge_data_dict:
                    edge_data_dict[k] = []

                edge_data_dict[k].append(v)

        edge_xml_list = []

        for edge_idx, (source_node_id, target_node_id, edge_data) in enumerate(
            graph.edges(data=True)
        ):
            edge_type_key = kwargs.get('edge_type_key', None)
            edge_type = edge_data.get(edge_type_key, '-')

            edge_label = self._create_edge_label(
                source_node_id, target_node_id, edge_type, edge_data, **kwargs
            )

            att_xml_list = self._create_edge_att_xml_list(
                source_node_id, target_node_id, edge_data, **kwargs
            )
            att_xml_list.extend(
                self._create_att_xml_list(
                    EDGE_ATT_ELEMENTS, edge_data, edge_data_dict, **kwargs
                )
            )

            graphics_xml = self._create_edge_graphics_xml(
                edge_data, edge_data_dict, **kwargs
            )

            edge_xml = E.edge(
                *att_xml_list,
                graphics_xml,
                {
                    'id': str(edge_idx),
                    'label': edge_label,
                    'source': source_node_id,
                    'target': target_node_id,
                    QName(NAMESPACES['cy'], 'directed'): get_directed_value(graph),
                },
            )

            edge_xml_list.append(edge_xml)

        return edge_xml_list

    def _calc_nodes_size(
        self,
        graph: nx.MultiGraph,
        **kwargs,
    ) -> dict[str, tuple[float, float]]:
        node_data_dict = self._create_node_data_dict(graph)

        nodes_size = {}

        for node_id, node_data in graph.nodes(data=True):
            height = self._create_value(
                'h',
                NODE_GRAPHICS_ATTS['h'],
                node_data,
                node_data_dict,
                **kwargs,
            )
            width = self._create_value(
                'w',
                NODE_GRAPHICS_ATTS['w'],
                node_data,
                node_data_dict,
                **kwargs,
            )

            nodes_size[node_id] = (float(height), float(width))

        return nodes_size

    def _calc_nodes_position(
        self, graph: nx.MultiGraph, graph_layout: Layout, graph_layout_kwargs, **kwargs
    ) -> dict[str, tuple[float, float]]:
        nodes_size = self._calc_nodes_size(graph, **kwargs)

        return graph_layout.layout(
            graph,
            nodes_size=nodes_size,
            **graph_layout_kwargs,
        )

    def _create_graph_xml(self, graph: nx.MultiGraph, **kwargs) -> ET.Element:
        graph_id = kwargs.get('graph_id', self._uuid)
        graph_label = kwargs.get('graph_label', self._uuid)
        graph_cy_document_version = kwargs.get('graph_cy_document_version', '3.0')

        graph_attributes = {
            'id': graph_id,
            'label': graph_label,
            'directed': get_directed_value(graph),
            QName(NAMESPACES['cy'], 'documentVersion'): graph_cy_document_version,
        }

        graph_layout_cls = kwargs.get('graph_layout', KamadaKawaiLayout)
        graph_layout_kwargs = kwargs.get('graph_layout_kwargs', {})
        graph_layout = graph_layout_cls(**graph_layout_kwargs)
        nodes_position = self._calc_nodes_position(
            graph, graph_layout, graph_layout_kwargs, **kwargs
        )

        graph_xml = self._element_maker.graph(
            *self._create_graph_att_xml_list(graph_layout, **kwargs),
            self._create_graph_graphics_xml(nodes_position, **kwargs),
            *self._create_node_xml_list(graph, nodes_position, **kwargs),
            *self._create_edge_xml_list(graph, **kwargs),
            graph_attributes,
        )

        return graph_xml

    def write(self, graph: T, xgmml_path: str, **kwargs):
        self._uuid = str(uuid.uuid4())

        graph = self._to_networkx(graph, **kwargs)
        graph_xml = self._create_graph_xml(graph, **kwargs)

        with open(xgmml_path, 'w') as f:
            f.write(
                ET.tostring(
                    graph_xml,
                    encoding='utf-8',
                    pretty_print=True,
                    standalone=True,
                    xml_declaration=True,
                ).decode('utf-8')
            )


__all__ = ['XGMMLWriter']



import networkx as nx
import igraph as ig

from xgmml.reader.xgmml_reader import XGMMLReader


class IgraphXGMMLReader(XGMMLReader[ig.Graph]):
    def __init__(self, **kwargs):
        super(IgraphXGMMLReader, self).__init__(**kwargs)

    def _networkx_to(self, graph: nx.MultiGraph, **kwargs) -> ig.Graph:
        return ig.Graph.from_networkx(graph)


__all__ = ['IgraphXGMMLReader']



import matplotlib as mpl


NODE_SHAPES = [
    'ELLIPSE',
    'RECTANGLE',
    'HEXAGON',
    'OCTAGON',
    'DIAMOND',
    'TRIANGLE',
    'ROUNDED_RECTANGLE',
]


LINE_TYPES = [
    'SOLID',
    'LONG_DASH',
    'DOT',
    'DASH_DOT',
    'EQUAL_DASH',
    'PARALLEL_LINES',
    'VERTICAL_SLASH',
    'FORWARD_SLASH',
    'BACKWARD_SLASH',
    'CONTIGUOUS_ARROW',
    'SEPARATE_ARROW',
    'SINEWAVE',
    'ZIGZAG',
]


ARROW_SHAPES = [
    'ARROW',
    'ARROW_SHORT',
    'CIRCLE',
    'CROSS_DELTA',
    'CROSS_OPEN_DELTA',
    'DELTA',
    'DELTA_SHORT_1',
    'DELTA_SHORT_2',
    'DIAMOND',
    'DIAMOND_SHORT_1',
    'DIAMOND_SHORT_2',
    'HALF_BOTTOM',
    'HALF_CIRCLE',
    'HALF_TOP',
    'NONE',
    'OPEN_CIRCLE',
    'OPEN_DELTA',
    'OPEN_DIAMOND',
    'OPEN_HALF_CIRCLE',
    'OPEN_SQUARE',
    'SQUARE',
    'T',
]


DEFAULT_FOREGROUND_COLOR_PALETTE = [
    '#1783FF',
    '#00C9C9',
    '#F0884D',
    '#D580FF',
    '#7863FF',
    '#60C42D',
    '#BD8F24',
    '#FF80CA',
    '#2491B3',
    '#17C76F',
]


def add_alpha(color_hex, alpha):
    r, g, b = mpl.colors.to_rgb(color_hex)

    return mpl.colors.to_hex(
        (
            r * alpha + 1 * (1 - alpha),
            g * alpha + 1 * (1 - alpha),
            b * alpha + 1 * (1 - alpha),
        )
    )


DEFAULT_BACKGROUND_COLOR_PALETTE = [
    add_alpha(color_hex, 0.2) for color_hex in DEFAULT_FOREGROUND_COLOR_PALETTE
]


__all__ = [
    'NODE_SHAPES',
    'LINE_TYPES',
    'ARROW_SHAPES',
    'DEFAULT_FOREGROUND_COLOR_PALETTE',
    'DEFAULT_BACKGROUND_COLOR_PALETTE',
    'add_alpha',
]



import sys
import networkx as nx

from abc import ABC, abstractmethod
from math import dist


class Layout(ABC):
    def __init__(
        self, name, canvas_width=1200, canvas_height=900, subgraph_spacing=100, **kwargs
    ):
        self._name = name
        self._canvas_width = canvas_width
        self._canvas_height = canvas_height
        self._subgraph_spacing = subgraph_spacing

    @property
    def name(self):
        return self._name

    def _calc_auto_scale(
        self,
        graph: nx.Graph,
        graph_layout: dict[str, tuple[float, float]],
        nodes_size: dict[str, tuple[float, float]],
        min_node_spacing_ratio=2.0,
    ) -> float:
        edge_node_distances = []
        for node_source, node_target in graph.edges():
            edge_node_distances.append(
                dist(graph_layout[node_source], graph_layout[node_target])
            )

        node_distances = []
        nodes = list(graph.nodes())
        for i in range(len(nodes)):
            for j in range(i + 1, len(nodes)):
                node_distances.append(
                    dist(graph_layout[nodes[i]], graph_layout[nodes[j]])
                )

        min_node_distance = max(min(edge_node_distances), 0.1)
        max_node_distance = max(node_distances)
        avg_node_size = sum([(w + h) / 2.0 for w, h in nodes_size.values()]) / len(
            nodes_size
        )

        return (
            min_node_spacing_ratio
            * avg_node_size
            * max_node_distance
            / min_node_distance
            / 2
        )

    def _get_graph_layout_stats(self, subgraph_layout: dict[str, tuple[float, float]]):
        node_x_list = [node_x for node_x, _ in subgraph_layout.values()]
        node_y_list = [node_y for _, node_y in subgraph_layout.values()]

        return {
            'min_x': min(node_x_list),
            'max_x': max(node_x_list),
            'min_y': min(node_y_list),
            'max_y': max(node_y_list),
            'width': max(node_x_list) - min(node_x_list),
            'height': max(node_y_list) - min(node_y_list),
        }

    def _calc_max_width(
        self, subgraphs_layout: list[dict[str, tuple[float, float]]], max_iterations=10
    ) -> float:
        subgraphs_layout_stats = [
            self._get_graph_layout_stats(subgraph_layout)
            for subgraph_layout in subgraphs_layout
        ]

        min_x = sys.float_info.max
        max_x = sys.float_info.min

        for subgraph_layout_stats in subgraphs_layout_stats:
            min_x = min(min_x, subgraph_layout_stats['min_x'])
            max_x = max(max_x, subgraph_layout_stats['max_x'])

        width = max_x - min_x
        max_width = max(width, self._canvas_width)
        canvas_ratio = self._canvas_width / self._canvas_height

        for _ in range(max_iterations):
            next_x_start = 0.0
            next_y_start = 0.0
            y_max = 0.0

            for subgraph_layout_stat in subgraphs_layout_stats:
                next_x_start += subgraph_layout_stat['width'] + self._subgraph_spacing
                y_max = max(y_max, subgraph_layout_stat['height'])

                if next_x_start > max_width:
                    next_x_start = 0.0
                    next_y_start += y_max + self._subgraph_spacing
                    y_max = 0.0

            layout_ratio = max_width / (next_y_start + y_max)
            max_width *= canvas_ratio / layout_ratio

        return max_width

    def _offset_subgraph_layout(
        self,
        subgraph_layout: dict[str, tuple[float, float]],
        subgraph_layout_stats: dict,
        offset_x,
        offset_y,
    ):
        min_x = subgraph_layout_stats['min_x']
        min_y = subgraph_layout_stats['min_y']

        for node_id, (node_x, node_y) in subgraph_layout.items():
            subgraph_layout[node_id] = (
                node_x + offset_x - min_x,
                node_y + offset_y - min_y,
            )

    def layout(
        self,
        graph: nx.MultiGraph,
        nodes_size: dict[str, tuple[float, float]] = None,
        **kwargs,
    ) -> dict[str, tuple[float, float]]:
        undirected_graph = graph.to_undirected()
        connected_components = list(nx.connected_components(undirected_graph))
        connected_components = sorted(
            connected_components, key=lambda x: len(x), reverse=True
        )
        subgraphs = [graph.subgraph(nodes).copy() for nodes in connected_components]

        subgraphs_layout = [
            self._subgraph_layout(subgraph, nodes_size) for subgraph in subgraphs
        ]

        max_width = self._calc_max_width(subgraphs_layout, **kwargs)
        next_x_start = 0.0
        next_y_start = 0.0
        y_max = 0.0

        for subgraph_layout in subgraphs_layout:
            subgraph_layout_stats = self._get_graph_layout_stats(subgraph_layout)
            self._offset_subgraph_layout(
                subgraph_layout, subgraph_layout_stats, next_x_start, next_y_start
            )

            next_x_start += subgraph_layout_stats['width'] + self._subgraph_spacing
            y_max = max(y_max, subgraph_layout_stats['height'])

            if next_x_start > max_width:
                next_x_start = 0.0
                next_y_start += y_max + self._subgraph_spacing
                y_max = 0.0

        layout = {}
        for subgraph_layout in subgraphs_layout:
            for node_id, node_position in subgraph_layout.items():
                layout[node_id] = node_position

        return layout

    @abstractmethod
    def _subgraph_layout(
        self,
        subgraph: nx.Graph,
        nodes_size: dict[str, tuple[float, float]] = None,
    ) -> dict[str, tuple[float, float]]:
        raise NotImplementedError()


__all__ = ['Layout']



import networkx as nx

from lxml.etree import QName

from xgmml.style.style_mapper import *


NAMESPACES = {
    'cy': 'http://www.cytoscape.org',
    'dc': 'http://purl.org/dc/elements/1.1/',
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'xlink': 'http://www.w3.org/1999/xlink',
    None: 'http://www.cs.rpi.edu/XGMML',
}


ATT_TYPES = {
    'boolean': {
        'type': 'boolean',
        QName(NAMESPACES['cy'], 'type'): 'Boolean',
    },
    'integer': {
        'type': 'integer',
        QName(NAMESPACES['cy'], 'type'): 'Integer',
    },
    'real': {
        'type': 'real',
        QName(NAMESPACES['cy'], 'type'): 'Double',
    },
    'string': {
        'type': 'string',
        QName(NAMESPACES['cy'], 'type'): 'String',
    },
}


GRAPH_GRAPHICS_ATT_ELEMENTS = {
    'NETWORK_ANNOTATION_SELECTION': {
        'kwarg': 'graph_annotation_selection',
        'default': 'false',
        'type': 'string',
    },
    'NETWORK_BACKGROUND_PAINT': {
        'kwarg': 'graph_background_color',
        'default': '#FFFFFF',
        'type': 'string',
    },
    'NETWORK_CENTER_X_LOCATION': {
        'kwarg': 'graph_center_x_location',
        'default': None,
        'type': 'string',
    },
    'NETWORK_CENTER_Y_LOCATION': {
        'kwarg': 'graph_center_y_location',
        'default': None,
        'type': 'string',
    },
    'NETWORK_CENTER_Z_LOCATION': {
        'kwarg': 'graph_center_z_location',
        'default': '0.0',
        'type': 'string',
    },
    'NETWORK_DEPTH': {
        'kwarg': 'graph_depth',
        'default': '0.0',
        'type': 'string',
    },
    'NETWORK_EDGE_SELECTION': {
        'kwarg': 'graph_edge_selection',
        'default': 'false',
        'type': 'string',
    },
    'NETWORK_FORCE_HIGH_DETAIL': {
        'kwarg': 'graph_force_high_detail',
        'default': 'false',
        'type': 'string',
    },
    'NETWORK_HEIGHT': {
        'kwarg': 'graph_height',
        'default': None,
        'type': 'string',
    },
    'NETWORK_NODE_LABEL_SELECTION': {
        'kwarg': 'graph_node_label_selection',
        'default': 'false',
        'type': 'string',
    },
    'NETWORK_NODE_SELECTION': {
        'kwarg': 'graph_node_selection',
        'default': 'true',
        'type': 'string',
    },
    'NETWORK_SCALE_FACTOR': {
        'kwarg': 'graph_scale_factor',
        'default': '1.0',
        'type': 'string',
    },
    'NETWORK_TITLE': {
        'kwarg': 'graph_label',
        'default': None,
        'type': 'string',
    },
    'NETWORK_WIDTH': {
        'kwarg': 'graph_width',
        'default': None,
        'type': 'string',
    },
}


NODE_ATTS = {
    'label': {
        'kwarg': 'node_label',
        'default': None,
        'type': 'string',
        'key_kwarg': 'node_label_key',
        'mapper_kwarg': 'node_label_mapper',
        'default_mapper': StylePassthroughMapper,
    }
}


NODE_ATT_ELEMENTS = {
    'selected': {
        'kwarg': 'node_selected',
        'default': '0',
        'type': 'boolean',
        'key_kwarg': 'node_selected_key',
        'mapper_kwarg': 'node_selected_mapper',
        'default_mapper': StylePassthroughMapper,
    },
}


NODE_GRAPHICS_ATTS = {
    'type': {
        'kwarg': 'node_shape',
        'default': 'ELLIPSE',
        'type': 'string',
        'key_kwarg': 'node_shape_key',
        'mapper_kwarg': 'node_shape_mapper',
        'default_mapper': StyleNodeShapeDefaultDiscreteMapper,
    },
    'outline': {
        'kwarg': 'node_outline_color',
        'default': '#CCCCCC',
        'type': 'string',
        'key_kwarg': 'node_outline_color_key',
        'mapper_kwarg': 'node_outline_color_mapper',
        'default_mapper': StyleForegroundColorDefaultDiscreteMapper,
    },
    'width': {
        'kwarg': 'node_outline_width',
        'default': '2.0',
        'type': 'string',
    },
    'fill': {
        'kwarg': 'node_fill_color',
        'default': '#FFFFFF',
        'type': 'string',
        'key_kwarg': 'node_fill_color_key',
        'mapper_kwarg': 'node_fill_color_mapper',
        'default_mapper': StyleBackgroundColorDefaultDiscreteMapper,
    },
    'h': {
        'kwarg': 'node_height',
        'default': '40.0',
        'type': 'string',
        'key_kwarg': 'node_height_key',
        'mapper_kwarg': 'node_height_mapper',
        'default_mapper': StyleNodeSizeDefaultContinuousMapper,
    },
    'w': {
        'kwarg': 'node_width',
        'default': '40.0',
        'type': 'string',
        'key_kwarg': 'node_width_key',
        'mapper_kwarg': 'node_width_mapper',
        'default_mapper': StyleNodeSizeDefaultContinuousMapper,
    },
}


NODE_GRAPHICS_ATT_ELEMENTS = {
    'COMPOUND_NODE_PADDING': {
        'kwarg': 'node_padding',
        'default': '10.0',
        'type': 'string',
    },
    'COMPOUND_NODE_SHAPE': {
        'kwarg': 'node_shape',
        'default': 'ELLIPSE',
        'type': 'string',
        'key_kwarg': 'node_shape_key',
        'mapper_kwarg': 'node_shape_mapper',
        'default_mapper': StyleNodeShapeDefaultDiscreteMapper,
    },
    'NODE_BORDER_STROKE': {
        'kwarg': 'node_border_stroke',
        'default': 'SOLID',
        'type': 'string',
    },
    'NODE_BORDER_TRANSPARENCY': {
        'kwarg': 'node_border_transparency',
        'default': '255',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_1': {
        'kwarg': 'node_custom_graphics_1',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_2': {
        'kwarg': 'node_custom_graphics_2',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_3': {
        'kwarg': 'node_custom_graphics_3',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_4': {
        'kwarg': 'node_custom_graphics_4',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_5': {
        'kwarg': 'node_custom_graphics_5',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_6': {
        'kwarg': 'node_custom_graphics_6',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_7': {
        'kwarg': 'node_custom_graphics_7',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_8': {
        'kwarg': 'node_custom_graphics_8',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_9': {
        'kwarg': 'node_custom_graphics_9',
        'default': 'org.cytoscape.cg.model.NullCustomGraphics,0,[ Remove Graphics ],',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_1': {
        'kwarg': 'node_custom_graphics_position_1',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_2': {
        'kwarg': 'node_custom_graphics_position_2',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_3': {
        'kwarg': 'node_custom_graphics_position_3',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_4': {
        'kwarg': 'node_custom_graphics_position_4',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_5': {
        'kwarg': 'node_custom_graphics_position_5',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_6': {
        'kwarg': 'node_custom_graphics_position_6',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_7': {
        'kwarg': 'node_custom_graphics_position_7',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_8': {
        'kwarg': 'node_custom_graphics_position_8',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_POSITION_9': {
        'kwarg': 'node_custom_graphics_position_9',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_1': {
        'kwarg': 'node_custom_graphics_size_1',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_2': {
        'kwarg': 'node_custom_graphics_size_2',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_3': {
        'kwarg': 'node_custom_graphics_size_3',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_4': {
        'kwarg': 'node_custom_graphics_size_4',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_5': {
        'kwarg': 'node_custom_graphics_size_5',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_6': {
        'kwarg': 'node_custom_graphics_size_6',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_7': {
        'kwarg': 'node_custom_graphics_size_7',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_8': {
        'kwarg': 'node_custom_graphics_size_8',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_CUSTOMGRAPHICS_SIZE_9': {
        'kwarg': 'node_custom_graphics_size_9',
        'default': '40.0',
        'type': 'string',
    },
    'NODE_DEPTH': {
        'kwarg': 'node_depth',
        'default': '0.0',
        'type': 'string',
    },
    'NODE_LABEL': {
        'kwarg': 'node_label',
        'default': None,
        'type': 'string',
        'key_kwarg': 'node_label_key',
        'mapper_kwarg': 'node_label_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'NODE_LABEL_BACKGROUND_COLOR': {
        'kwarg': 'node_label_background_color',
        'default': '#B6B6B6',
        'type': 'string',
    },
    'NODE_LABEL_BACKGROUND_SHAPE': {
        'kwarg': 'node_label_background_shape',
        'default': 'NONE',
        'type': 'string',
    },
    'NODE_LABEL_BACKGROUND_TRANSPARENCY': {
        'kwarg': 'node_label_background_transparency',
        'default': '255',
        'type': 'string',
    },
    'NODE_LABEL_COLOR': {
        'kwarg': 'node_label_color',
        'default': '#333333',
        'type': 'string',
    },
    'NODE_LABEL_FONT_FACE': {
        'kwarg': 'node_label_font_face',
        'default': 'Dialog,plain,12',
        'type': 'string',
    },
    'NODE_LABEL_FONT_SIZE': {
        'kwarg': 'node_label_font_size',
        'default': '10',
        'type': 'string',
        'key_kwarg': 'node_label_font_size_key',
        'mapper_kwarg': 'node_label_font_size_mapper',
        'default_mapper': StyleFontSizeDefaultContinuousMapper,
    },
    'NODE_LABEL_POSITION': {
        'kwarg': 'node_label_position',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'NODE_LABEL_ROTATION': {
        'kwarg': 'node_label_rotation',
        'default': '0.0',
        'type': 'string',
    },
    'NODE_LABEL_TRANSPARENCY': {
        'kwarg': 'node_label_transparency',
        'default': '255',
        'type': 'string',
    },
    'NODE_LABEL_WIDTH': {
        'kwarg': 'node_label_width',
        'default': '200.0',
        'type': 'string',
    },
    'NODE_NESTED_NETWORK_IMAGE_VISIBLE': {
        'kwarg': 'node_nested_network_image_visible',
        'default': 'true',
        'type': 'string',
    },
    'NODE_SELECTED': {
        'kwarg': 'node_selected',
        'default': 'false',
        'type': 'string',
        'key_kwarg': 'node_selected_key',
        'mapper_kwarg': 'node_selected_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'NODE_SELECTED_PAINT': {
        'kwarg': 'node_selected_color',
        'default': '#FFFF00',
        'type': 'string',
    },
    'NODE_TOOLTIP': {
        'kwarg': 'node_tooltip',
        'default': '',
        'type': 'string',
        'key_kwarg': 'node_tooltip_key',
        'mapper_kwarg': 'node_tooltip_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'NODE_TRANSPARENCY': {
        'kwarg': 'node_transparency',
        'default': '255',
        'type': 'string',
    },
    'NODE_VISIBLE': {
        'kwarg': 'node_visible',
        'default': 'true',
        'type': 'string',
        'key_kwarg': 'node_visible_key',
        'mapper_kwarg': 'node_visible_mapper',
        'default_mapper': StylePassthroughMapper,
    },
}


EDGE_ATTS = {}


EDGE_ATT_ELEMENTS = {
    'selected': {
        'kwarg': 'edge_selected',
        'default': 'false',
        'type': 'boolean',
        'key_kwarg': 'edge_selected_key',
        'mapper_kwarg': 'edge_selected_mapper',
        'default_mapper': StylePassthroughMapper,
    },
}


EDGE_GRAPHICS_ATTS = {
    'width': {
        'kwarg': 'edge_width',
        'default': '2.0',
        'type': 'string',
    },
    'fill': {
        'kwarg': 'edge_fill_color',
        'default': '#999999',
        'type': 'string',
        'key_kwarg': 'edge_fill_color_key',
        'mapper_kwarg': 'edge_fill_color_mapper',
        'default_mapper': StyleBackgroundColorDefaultDiscreteMapper,
    },
}


EDGE_GRAPHICS_ATT_ELEMENTS = {
    'EDGE_BEND': {
        'kwarg': 'edge_bend',
        'default': '',
        'type': 'string',
    },
    'EDGE_CURVED': {
        'kwarg': 'edge_curved',
        'default': 'true',
        'type': 'string',
    },
    'EDGE_LABEL': {
        'kwarg': 'edge_label',
        'default': '',
        'type': 'string',
        'key_kwarg': 'edge_label_key',
        'mapper_kwarg': 'edge_label_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'EDGE_LABEL_AUTOROTATE': {
        'kwarg': 'edge_label_autorotate',
        'default': 'false',
        'type': 'string',
    },
    'EDGE_LABEL_BACKGROUND_COLOR': {
        'kwarg': 'edge_label_background_color',
        'default': '#B6B6B6',
        'type': 'string',
    },
    'EDGE_LABEL_BACKGROUND_SHAPE': {
        'kwarg': 'edge_label_background_shape',
        'default': 'NONE',
        'type': 'string',
    },
    'EDGE_LABEL_BACKGROUND_TRANSPARENCY': {
        'kwarg': 'edge_label_background_transparency',
        'default': '255',
        'type': 'string',
    },
    'EDGE_LABEL_COLOR': {
        'kwarg': 'edge_label_color',
        'default': '#000000',
        'type': 'string',
    },
    'EDGE_LABEL_FONT_FACE': {
        'kwarg': 'edge_label_font_face',
        'default': 'Dialog,plain,10',
        'type': 'string',
    },
    'EDGE_LABEL_FONT_SIZE': {
        'kwarg': 'edge_label_font_size',
        'default': '10',
        'type': 'string',
    },
    'EDGE_LABEL_POSITION': {
        'kwarg': 'edge_label_position',
        'default': 'C,C,c,0.00,0.00',
        'type': 'string',
    },
    'EDGE_LABEL_ROTATION': {
        'kwarg': 'edge_label_rotation',
        'default': '0.0',
        'type': 'string',
    },
    'EDGE_LABEL_TRANSPARENCY': {
        'kwarg': 'edge_label_transparency',
        'default': '255',
        'type': 'string',
    },
    'EDGE_LABEL_WIDTH': {
        'kwarg': 'edge_label_width',
        'default': '200.0',
        'type': 'string',
    },
    'EDGE_LINE_TYPE': {
        'kwarg': 'edge_line_type',
        'default': 'SOLID',
        'type': 'string',
        'key_kwarg': 'edge_line_type_key',
        'mapper_kwarg': 'edge_line_type_mapper',
        'default_mapper': StyleLineTypeDefaultDiscreteMapper,
    },
    'EDGE_SELECTED': {
        'kwarg': 'edge_selected',
        'default': 'false',
        'type': 'string',
        'key_kwarg': 'edge_selected_key',
        'mapper_kwarg': 'edge_selected_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'EDGE_SOURCE_ARROW_SELECTED_PAINT': {
        'kwarg': 'edge_source_arrow_selected_color',
        'default': '#FFFF00',
        'type': 'string',
    },
    'EDGE_SOURCE_ARROW_SHAPE': {
        'kwarg': 'edge_source_arrow_shape',
        'default': 'NONE',
        'type': 'string',
        'key_kwarg': 'edge_source_arrow_shape_key',
        'mapper_kwarg': 'edge_source_arrow_shape_mapper',
        'default_mapper': StyleArrowShapeDefaultDiscreteMapper,
    },
    'EDGE_SOURCE_ARROW_SIZE': {
        'kwarg': 'edge_source_arrow_size',
        'default': '6.0',
        'type': 'string',
    },
    'EDGE_SOURCE_ARROW_UNSELECTED_PAINT': {
        'kwarg': 'edge_source_arrow_unselected_color',
        'default': '#000000',
        'type': 'string',
    },
    'EDGE_STACKING': {
        'kwarg': 'edge_stacking',
        'default': 'AUTO_BEND',
        'type': 'string',
    },
    'EDGE_STACKING_DENSITY': {
        'kwarg': 'edge_stacking_density',
        'default': '0.5',
        'type': 'string',
    },
    'EDGE_STROKE_SELECTED_PAINT': {
        'kwarg': 'edge_stroke_selected_color',
        'default': '#FF0000',
        'type': 'string',
    },
    'EDGE_TARGET_ARROW_SELECTED_PAINT': {
        'kwarg': 'edge_target_arrow_selected_color',
        'default': '#FFFF00',
        'type': 'string',
    },
    'EDGE_TARGET_ARROW_SHAPE': {
        'kwarg': 'edge_target_arrow_shape',
        'default': 'NONE',
        'type': 'string',
        'key_kwarg': 'edge_target_arrow_shape_key',
        'mapper_kwarg': 'edge_target_arrow_shape_mapper',
        'default_mapper': StyleArrowShapeDefaultDiscreteMapper,
    },
    'EDGE_TARGET_ARROW_SIZE': {
        'kwarg': 'edge_target_arrow_size',
        'default': '6.0',
        'type': 'string',
    },
    'EDGE_TARGET_ARROW_UNSELECTED_PAINT': {
        'kwarg': 'edge_target_arrow_unselected_color',
        'default': '#000000',
        'type': 'string',
    },
    'EDGE_TOOLTIP': {
        'kwarg': 'edge_tooltip',
        'default': '',
        'type': 'string',
        'key_kwarg': 'edge_tooltip_key',
        'mapper_kwarg': 'edge_tooltip_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'EDGE_TRANSPARENCY': {
        'kwarg': 'edge_transparency',
        'default': '170',
        'type': 'string',
    },
    'EDGE_VISIBLE': {
        'kwarg': 'edge_visible',
        'default': 'true',
        'type': 'string',
        'key_kwarg': 'edge_visible_key',
        'mapper_kwarg': 'edge_visible_mapper',
        'default_mapper': StylePassthroughMapper,
    },
    'EDGE_Z_ORDER': {
        'kwarg': 'edge_z_order',
        'default': '0.0',
        'type': 'string',
    },
}


IGNORED_NODE_ATT_ELEMENTS = [
    'name',
    'selected',
    'shared name',
]


IGNORED_EDGE_ATT_ELEMENTS = {
    'id',
    'interaction',
    'label',
    'name',
    'selected',
    'shared interaction',
    'shared name',
    'source',
    'target',
    str(QName(NAMESPACES['cy'], 'directed')),
}


def get_att_value_type(value: any) -> str:
    if isinstance(value, bool):
        return 'boolean'
    elif isinstance(value, int):
        return 'integer'
    elif isinstance(value, float):
        return 'real'
    else:
        return 'string'


def get_att_value(value, value_type='string'):
    if value_type == 'boolean':
        return bool(int(value))
    elif value_type == 'integer':
        return int(value)
    elif value_type == 'real':
        return float(value)
    else:
        return str(value)


def get_directed_value(graph: nx.MultiGraph) -> str:
    return '1' if graph.is_directed() else '0'


def create_att_value(value, value_type='string'):
    if value is None:
        return None

    if isinstance(value, bool):
        if value_type == 'boolean':
            return str(int(value))
        else:
            return str(value).lower()
    else:
        return str(value)


def create_att_name_value(name, value, value_type='string') -> dict:
    att = {
        'name': name,
    }

    value = create_att_value(value, value_type=value_type)

    if value is not None:
        att['value'] = value

    if value_type in ATT_TYPES:
        att.update(ATT_TYPES[value_type])

    return att


def first_or_none(values):
    return values[0] if len(values) > 0 else None


__all__ = [
    'NAMESPACES',
    'ATT_TYPES',
    'GRAPH_GRAPHICS_ATT_ELEMENTS',
    'NODE_ATTS',
    'NODE_ATT_ELEMENTS',
    'NODE_GRAPHICS_ATTS',
    'NODE_GRAPHICS_ATT_ELEMENTS',
    'EDGE_ATTS',
    'EDGE_ATT_ELEMENTS',
    'EDGE_GRAPHICS_ATTS',
    'EDGE_GRAPHICS_ATT_ELEMENTS',
    'IGNORED_NODE_ATT_ELEMENTS',
    'IGNORED_EDGE_ATT_ELEMENTS',
    'get_att_value_type',
    'get_att_value',
    'get_directed_value',
    'create_att_value',
    'create_att_name_value',
    'first_or_none',
]



import matplotlib as mpl

from abc import ABC, abstractmethod

from xgmml.utils.style_utils import *


class StyleMapper(ABC):
    def __init__(self, values: list, **kwargs):
        self._values = values

    @abstractmethod
    def map(self, value):
        raise NotImplementedError()


class StylePassthroughMapper(StyleMapper):
    def __init__(self, values: list, **kwargs):
        super(StylePassthroughMapper, self).__init__(values)

    def map(self, value):
        return value


class StyleDiscreteMapper(StyleMapper):
    def __init__(self, values: list, styles: list, **kwargs):
        super(StyleDiscreteMapper, self).__init__(values)

        self._styles = styles
        self._categories = sorted(set(values), key=self._sort_key)

    def _sort_key(self, value):
        return value

    def map(self, value):
        category_idx = self._categories.index(value)
        style_idx = category_idx % len(self._styles)

        return self._styles[style_idx]


class StyleContinuousMapper(StyleMapper):
    def __init__(self, values: list, **kwargs):
        super(StyleContinuousMapper, self).__init__(values)

        self._min_value = min(self._values)
        self._max_value = max(self._values)

    def _transform_value(self, value) -> float:
        if self._max_value - self._min_value == 0:
            return 0.0

        return (value - self._min_value) / (self._max_value - self._min_value)

    def map(self, value):
        return self._transform_value(value)


class StyleColorDiscreteMapper(StyleMapper):
    def __init__(self, values: list, colormap: mpl.colors.ListedColormap, **kwargs):
        super(StyleColorDiscreteMapper, self).__init__(values)

        self._colors = colormap.colors
        self._categories = sorted(set(values), key=self._sort_key)

    def _sort_key(self, value):
        return value

    def map(self, value):
        category_idx = self._categories.index(value)
        color_idx = category_idx % len(self._colors)

        return mpl.colors.to_hex(self._colors[color_idx])


class StyleColorContinuousMapper(StyleMapper):
    def __init__(
        self, values: list, colormap: mpl.colors.LinearSegmentedColormap, **kwargs
    ):
        super(StyleColorContinuousMapper, self).__init__(values)

        self._colormap = colormap
        self._normalizer = mpl.colors.Normalize(vmin=min(values), vmax=max(values))

    def map(self, value):
        color = self._colormap(self._normalizer(value))

        return mpl.colors.to_hex(color).upper()


class StyleNodeShapeDefaultDiscreteMapper(StyleDiscreteMapper):
    def __init__(self, values, **kwargs):
        super(StyleNodeShapeDefaultDiscreteMapper, self).__init__(
            values, NODE_SHAPES, **kwargs
        )


class StyleNodeSizeDefaultContinuousMapper(StyleContinuousMapper):
    def __init__(self, values, **kwargs):
        super(StyleNodeSizeDefaultContinuousMapper, self).__init__(values, **kwargs)

        self._min_size = 10
        self._max_size = 40

    def _transform_value(self, value):
        transformed_value = super(
            StyleNodeSizeDefaultContinuousMapper, self
        )._transform_value(value)

        return int(
            transformed_value * (self._max_size - self._min_size) + self._min_size
        )


class StyleForegroundColorDefaultDiscreteMapper(StyleColorDiscreteMapper):
    def __init__(self, values, **kwargs):
        super(StyleForegroundColorDefaultDiscreteMapper, self).__init__(
            values,
            mpl.colors.ListedColormap(DEFAULT_FOREGROUND_COLOR_PALETTE),
            **kwargs,
        )


class StyleBackgroundColorDefaultDiscreteMapper(StyleColorDiscreteMapper):
    def __init__(self, values, **kwargs):
        super(StyleBackgroundColorDefaultDiscreteMapper, self).__init__(
            values,
            mpl.colors.ListedColormap(DEFAULT_BACKGROUND_COLOR_PALETTE),
            **kwargs,
        )


class StyleColorDefaultContinuousMapper(StyleColorContinuousMapper):
    def __init__(self, values, **kwargs):
        super(StyleColorDefaultContinuousMapper, self).__init__(
            values, mpl.colormaps['coolwarm'], **kwargs
        )


class StyleFontSizeDefaultContinuousMapper(StyleContinuousMapper):
    def __init__(self, values, **kwargs):
        super(StyleFontSizeDefaultContinuousMapper, self).__init__(values, **kwargs)

        self._min_size = 10
        self._max_size = 40

    def _transform_value(self, value):
        transformed_value = super(
            StyleFontSizeDefaultContinuousMapper, self
        )._transform_value(value)

        return int(
            transformed_value * (self._max_size - self._min_size) + self._min_size
        )


class StyleLineTypeDefaultDiscreteMapper(StyleDiscreteMapper):
    def __init__(self, values, **kwargs):
        super(StyleLineTypeDefaultDiscreteMapper, self).__init__(
            values, LINE_TYPES, **kwargs
        )


class StyleArrowShapeDefaultDiscreteMapper(StyleDiscreteMapper):
    def __init__(self, values, **kwargs):
        super(StyleArrowShapeDefaultDiscreteMapper, self).__init__(
            values, ARROW_SHAPES, **kwargs
        )


__all__ = [
    'StyleMapper',
    'StylePassthroughMapper',
    'StyleDiscreteMapper',
    'StyleContinuousMapper',
    'StyleColorDiscreteMapper',
    'StyleColorContinuousMapper',
    'StyleNodeShapeDefaultDiscreteMapper',
    'StyleNodeSizeDefaultContinuousMapper',
    'StyleForegroundColorDefaultDiscreteMapper',
    'StyleBackgroundColorDefaultDiscreteMapper',
    'StyleColorDefaultContinuousMapper',
    'StyleFontSizeDefaultContinuousMapper',
    'StyleLineTypeDefaultDiscreteMapper',
    'StyleArrowShapeDefaultDiscreteMapper',
]



import networkx as nx
import igraph as ig

from xgmml.writer.xgmml_writer import XGMMLWriter


class IgraphXGMMLWriter(XGMMLWriter[ig.Graph]):
    def __init__(self, **kwargs):
        super(IgraphXGMMLWriter, self).__init__(**kwargs)

    def _to_networkx(self, graph: ig.Graph, **kwargs) -> nx.MultiGraph:
        return ig.Graph.to_networkx(graph)


__all__ = ['IgraphXGMMLWriter']



import networkx as nx

from xgmml.writer.xgmml_writer import XGMMLWriter


class NetworkxXGMMLWriter(XGMMLWriter[nx.MultiGraph]):
    def __init__(self, **kwargs):
        super(NetworkxXGMMLWriter, self).__init__(**kwargs)

    def _to_networkx(self, graph: nx.MultiGraph, **kwargs) -> nx.MultiGraph:
        return graph


__all__ = ['NetworkxXGMMLWriter']
